/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef lint
static char sccsid[] = "@(#)$RCSfile: lcg-cp.c,v $ $Revision: 1.20 $ $Date: 2009/11/18 13:53:33 $ CERN Jean-Philippe Baud";
#endif /* not lint */

/*	lcg-cp - copy a Grid file to a local destination */
#include <sys/types.h>
#include <errno.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "lcg_util.h"

main(int argc, char **argv)
{
	int helpflg = 0;
	int insecure = 0;
	int verbose = 0;
	int version = 0;
	int c;
	char *conf_file = NULL;
	char *dp;
	int errflg = 0;
	char** src_protocols = NULL;
	char** dst_protocols = NULL;
	int res = -1;

	static struct option longopts[] = {
		{"config", required_argument, NULL, 'c'},
		{"help", no_argument, NULL, 'h'},
		{"insecure", no_argument, NULL, 'i'},
		{"sst", required_argument, NULL, 's'},
		{"dst", required_argument, NULL, 'S'},
		{"defaultsetype", required_argument, NULL, 'D'},
		{"srcsetype", required_argument, NULL, 'T'},
		{"dstsetype", required_argument, NULL, 'U'},
		{"nobdii", no_argument, NULL, 'b'},
		{"verbose", no_argument, NULL, 'v'},
		{"checksum", no_argument, NULL, 'k'},
		{"checksum-type", required_argument, NULL, 'K'},
		{"vo", required_argument, NULL, 'V'},
		{"timeout", required_argument, NULL, 't'},
#define CONNECT_TIMEOUT_ARG_ID 130
		{"connect-timeout", required_argument, NULL, CONNECT_TIMEOUT_ARG_ID},
#define SENDRECEIVE_TIMEOUT_ARG_ID 131
		{"sendreceive-timeout", required_argument, NULL, SENDRECEIVE_TIMEOUT_ARG_ID},
#define BDII_TIMEOUT_ARG_ID 132
		{"bdii-timeout", required_argument, NULL, BDII_TIMEOUT_ARG_ID},
#define SRM_TIMEOUT_ARG_ID 133
		{"srm-timeout", required_argument, NULL, SRM_TIMEOUT_ARG_ID},
#define SRC_PROTOCOLS_ARG_ID 134
	    {"src-protocols", required_argument, NULL, SRC_PROTOCOLS_ARG_ID},
#define DST_PROTOCOLS_ARG_ID 135
	    {"dst-protocols", required_argument, NULL, DST_PROTOCOLS_ARG_ID},
		{"version", no_argument, NULL, 'W'},
		{0, 0, 0, 0}
	};
	int nbstreams = 1;
	int timeout = 0, timeout_tmp = 0;
	char *vo = NULL;
	char *src_spacetokendesc = NULL;
	char *dest_spacetokendesc = NULL;
	enum se_type defaulttype = TYPE_NONE, src_type = TYPE_NONE, dst_type = TYPE_NONE;
	enum gfal_cksm_type cksmtype = GFAL_CKSM_NONE;
	int nobdii = 0;

	while ((c = getopt_long (argc, argv, "c::hkK:n:is:S:D:T:U:bvV:t:W", longopts, NULL)) != EOF) {
		switch (c) {
			case 'c':
				conf_file = optarg;
				break;
			case 'h':
				helpflg = 1;
				break;
			case 'i':
				insecure = 1;
				break;
			case 'n':
				if ((nbstreams = strtol (optarg, &dp, 10)) < 1 ||
						*dp != '\0') {
					fprintf (stderr,
							"invalid number of streams %s\n", optarg);
					errflg++;
				}
				break;
			case 's':
				src_spacetokendesc = optarg;
				break;
			case 'S':
				dest_spacetokendesc = optarg;
				break;
			case 'D':
				if (!strcasecmp (optarg, "srmv1"))
					defaulttype = TYPE_SRM;
				else if (!strcasecmp (optarg, "srmv2"))
					defaulttype = TYPE_SRMv2;
				else if (!strcasecmp (optarg, "se"))
					defaulttype = TYPE_SE;
				else {
					fprintf (stderr,"Invalid SE type: %s\n", optarg);
				}
				break;
			case 'T':
				if (!strcmp (optarg, "srmv1"))
					src_type = TYPE_SRM;
				else if (!strcmp (optarg, "srmv2"))
					src_type = TYPE_SRMv2;
				else if (!strcmp (optarg, "se"))
					src_type = TYPE_SE;
				else {
					fprintf (stderr,"Invalid SE type: %s\n", optarg);
					errflg++;
				}
				break;
			case 'U':
				if (!strcmp (optarg, "srmv1"))
					dst_type = TYPE_SRM;
				else if (!strcmp (optarg, "srmv2"))
					dst_type = TYPE_SRMv2;
				else if (!strcmp (optarg, "se"))
					dst_type = TYPE_SE;
				else {
					fprintf (stderr,"Invalid SE type: %s\n", optarg);
					errflg++;
				}
				break;
			case 'b':
				nobdii = 1;
				break;
			case 'V':
				vo = optarg;
				break;
			case 'v':
				++verbose;
				break;
			case 'k':
				if (cksmtype == GFAL_CKSM_NONE)
					cksmtype = GFAL_CKSM_DEFAULT;
				break;
			case 'K':
				cksmtype = gfal_parse_checksum_algname (optarg);
				if (cksmtype == GFAL_CKSM_NONE) {
					fprintf (stderr,"Invalid checksum type: %s\n", optarg);
				}
				break;
			case 't':
				fprintf (stderr, "Warning: -t,--timeout is deprecated! Use --timeout-* options instead\n");
				if ((timeout = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--timeout': %s\n", optarg);
					errflg++;
				}
			case CONNECT_TIMEOUT_ARG_ID: // "connect-timeout"
				if ((timeout_tmp = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--connect-timeout': %s\n", optarg);
					errflg++;
				} else
					gfal_set_timeout_connect (timeout_tmp);
				break;
			case SENDRECEIVE_TIMEOUT_ARG_ID: // "sendreceive-timeout"
				if ((timeout_tmp = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--sendreceive-timeout': %s\n", optarg);
					errflg++;
				} else
					gfal_set_timeout_sendreceive (timeout_tmp);
				break;
			case BDII_TIMEOUT_ARG_ID: // "bdii-timeout"
				if ((timeout_tmp = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--bdii-timeout': %s\n", optarg);
					errflg++;
				} else
					gfal_set_timeout_bdii (timeout_tmp);
				break;
			case SRM_TIMEOUT_ARG_ID: // "srm-timeout"
				if ((timeout_tmp = strtol (optarg, &dp, 10)) < 0 || 
						*dp != '\0' ) {
					fprintf (stderr,
							"invalid value for '--srm-timeout': %s\n", optarg);
					errflg++;
				} else
					gfal_set_timeout_srm (timeout_tmp);
				break;
            case SRC_PROTOCOLS_ARG_ID:
                src_protocols = lcg_tokenize(optarg, PROTOCOL_LIST_DELIMITER);

                if (!src_protocols) {
                    fprintf (stderr, "invalid source protocol list format: %s\n", optarg);
                    errflg++;
                }
                break;
            case DST_PROTOCOLS_ARG_ID:
                dst_protocols = lcg_tokenize(optarg, PROTOCOL_LIST_DELIMITER);

                if (!dst_protocols) {
                    fprintf (stderr, "invalid destination protocol list format: %s\n", optarg);
                    errflg++;
                }
                break;
			case 'W':
				version = 1;
				break;
			default:
				errflg++;
				break;
		}
	}
	if (version) {
		printf ("lcg_util-%s\n", lcg_util_version ());
		printf ("GFAL-client-%s\n", gfal_version ());
		exit (0);
	}
	if (optind + 2 != argc) {
		errflg++;
	}
	if (errflg || helpflg) {
		fprintf (stderr, "usage: %s %s%s%s%s%s", argv[0],
				"[-h,--help] [-c,--config config_file] [-n nbstreams] [-b,--nobdii] [-D,--defaultsetype se|srmv1|srmv2]\n",
				"\t[-T,--srcsetype se|srmv1|srmv2] [-U,--dstsetype se|srmv1|srmv2] [-s,--sst src_spacetokendesc]\n",
			    "\t[-S,--dst dest_spacetokendesc] [--connect-timeout value] [--sendreceive-timeout value]\n",
				"\t[--bdii-timeout value] [--srm-timeout value] [--checksum] [--checksum-type type]\n",
				"\t[-v,--verbose] [-V,--vo vo] [--version]\n"
				"\t[--src-protocols <protocol1,protocol2,...>]\n"
				"\t[--dst-protocols <protocol1,protocol2,...>]\n"
				"\tsrc_file  dest_file\n");
		exit (errflg ? 1 : 0);
	}

	res = lcg_cp5 (argv[optind], argv[optind + 1], defaulttype, src_type,
	    dst_type, nobdii, vo, nbstreams, conf_file, insecure, verbose, timeout,
	    src_spacetokendesc, dest_spacetokendesc, cksmtype, src_protocols,
	    dst_protocols, NULL, 0);

	lcg_free_string_array(src_protocols);
	lcg_free_string_array(dst_protocols);

	if (res < 0) {
		if (errno){
			perror ("lcg_cp");
		}
		exit (1);
	}
	exit (0);
}
