/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef lint
static char sccsid[] = "@(#)$RCSfile: lcg-replica-manager.c,v $ $Revision: 1.14 $ $Date: 2008/11/11 15:42:22 $ CERN James Casey";
#endif /* not lint */

/*	lcg-replica-manager - A replacement for the edg-replica-manager
 * program which delegates to lcg-utils commands
 */

#include <sys/types.h>
#include <errno.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

char *usage[] = {
  "-h,--help                       print help (if command is",
  "                                given, details on command)",
  "-i,--insecure                   Connect in an insecure",
  "                                manner, i.e. not https",
  "-v,--verbose                    print additional information",
  "                                while executing",
  "",
  "Commands:",
  "",
  "Management commands",
  "",
  "  copyAndRegisterFile   cr,creg Put a local file into Grid Storage and",
  "                                register it in the Catalog",
  "  replicateFile         rep     Replicate an existing file to a certain Grid",
  "                                Storage, update Catalog",
  "  deleteFile            del     Delete a file from Storage and remove entry",
  "                                from Catalog",
  "  getTurl               gt      Get a TURL given a SURL and a protocol.",
  "",
  "Catalog commands",
  "",
  "  registerFile          rf      Register a file in the catalog",
  "  unregisterFile        uf      Unregister a file from the catalog",
  "  listReplicas          lr      List all replicas of a logical file name",
  "  listGUID              lg      List the GUID of a known LFN or SURL",
  "  addAlias              aa      Add a new alias to GUID mapping",
  "  removeAlias           ra      remove an alias to GUID mapping",
  "",
  "",
  "File transfer commands",
  "",
  "  copyFile              cp      Copy a file. Use this to copy a file to local",
  "                                store by specifying an LFN as the source and a",
  "                                local destination.",
  NULL
};

int print_usage(const char *command) {
  int i;
  fprintf (stderr, "usage : %s %s \n", command, "[options] command [command-options]");
    for(i = 0; usage[i] != NULL; i++) {
      fprintf (stderr, "%s\n", usage[i]);
    }

}

int errflg = 0;
int helpflg;
char **args;
int size;
int idx;

int
add_optarg(const char *arg1, char *arg2) {
  char **p;
  char *str;

  if((str = strdup (arg1)) == NULL) {
    perror ("add_optarg");
    errflg++;
    return (-1);
  }
  if(idx + 2 >= size) {
    size *= 2;
    if((p = (char**) realloc (args, size * sizeof(char*))) == NULL) {
      free (args);
      perror ("add_optarg");
      errflg++;
      return (-1);
    }
    args = p;
  }
  args[idx++] = str;
  args[idx++] = arg2;
  return (0);
}

int
add_opt(const char *arg) {
  char *str;
  char **p;

  if((str = strdup (arg)) == NULL) {
    perror ("add_opt");
    errflg++;
    return (-1);
  }
  if(idx + 1 >= size) {
    size *= 2;
    if((p = (char**) realloc (args, size * sizeof(char*))) == NULL) {
      free (args);
      perror ("add_opt");
      errflg++;
      return (-1);
    }
    args = p;
  }
  args[idx++] = str;
  return (0);
}

int verbose;
main(argc, argv)
int argc;
char**argv;
{
  int c;

  char **p;
  char cmd[32];
  int i;
  char se_env[64];
  char *default_se;
  int has_dest = 0;
  char *vo_arg = NULL;

  /*
    We take the union of all accepted edg-rm options, filter out the
    not-supported ones and pass the rest on to the appropriate underlying 
    lcg_utils command.  

    Generally, the lcg-* tools only accept short options for many of these, so we map down the
    options.  All the edg-rm options are shown below
    
    ACCEPTED EDG-RM OPTIONS

-a,--all-available             (both map to -a)
   --config <file>
-d,--destination-file <file>   (both map to -d)
-h,--help
-i,--insecure
-l,--logical-file-name <file>  (both map to -l)
-n,--streams <#streams>        (both map to -n)
-s,--storage <host>            (both map to -s)
-v,--verbose
   --vo <VO>

   IGNORED EDG-RM OPTIONS
(all)
   --all-lfn                    (edg-rm del)
-f,--force                      (edg-rm cp)
-l,--localonly                  (edg-rm lr)
   --log-debug
   --log-info
   --log-off
-p,--protocol <protocol>        (edg-rep, cp, cr)
   --timing <file>
   --version




    NOT ACCEPTED EDG-RM OPTIONS (FROM ROS COMMANDS)
-d,--computing-elements <ceID>  (edg-rm ac)
-l,--logical-file-names <file>  (edg-rm ac)
-t,--turlProtocol <protocol>    (edg-rm gbf/lbf/ac)


*/
  static struct option longopts[] = {
    {"all-available", no_argument, 0, 'a'},
    {"all-lfn", no_argument, 0, 1}, /* no */
    {"config", required_argument, 0, 'C'}, 
    {"destination-file", required_argument, 0, 'd'},
    {"force", no_argument, 0, 1}, /* no */
    {"help", no_argument, &helpflg, 'h'},
    {"insecure", no_argument, 0, 'i'},
    {"logical-file-name", required_argument, 0, 'l'},
    {"localonly", no_argument, 0, 1}, /* no */
    {"log-debug", no_argument, 0, 1}, /* no */
    {"log-info", no_argument, 0, 1}, /* no */
    {"log-off", no_argument, 0, 1}, /* no */
    {"protocol", no_argument, 0, 1}, /* no */
    {"streams", required_argument, 0, 'n'},
    {"storage", required_argument, 0, 's'},
    {"timing", required_argument, 0, 1}, /* no */
    {"verbose", no_argument, &verbose, 'v'},
    {"version", no_argument, 0, 1}, /* no */
    {"vo", required_argument, 0, 'V'},
    {0, 0, 0, 0}
  };

  /* initially, set size to be argc, which should be big enough, unless
     people do things like '-hiv' */
  size = argc;

  if((args = (char**)calloc (size, sizeof(char*))) == NULL) {
    perror ("lcg-replica-manager");
    exit (1);
  }

  /* leave room for command name */
  idx = 1;
  opterr = 0;
  while ((c = getopt_long (argc, argv, "aAC:d:g:hil:n:p:P:s:vV:t:", longopts, NULL)) != EOF) {
    switch (c) {
    case 'a' :
      add_opt ("-a"); 
      break;
    case 'A' :
      add_opt ("-A");
      break;
    case 'C' :
      add_optarg ("--config", optarg);
      break;
    case 'd':
      add_optarg ("-d", optarg);
      has_dest = 1;
      break;
    case 'g':
      add_optarg ("-g", optarg);
      break;
    case 'h':
      add_opt ("-h");
      helpflg = 1;
      break;
    case 'i':
      add_opt ("-i");
      break;
    case 'l':
      add_optarg ("-l", optarg);
      break;
    case 'n':
      add_optarg ("-n", optarg);
      break;
    case 'P':
      add_optarg ("-P", optarg);
      break;
    case 's':
      add_optarg ("-s", optarg); 
    case 'v':
      add_opt ("-v");
      ++verbose;
      break;
    case 'V':
      vo_arg = optarg;
      add_optarg ("--vo", optarg); 
      break;
    case 't':
      vo_arg = optarg;
      add_optarg ("-t", optarg);
      break;

    case '?':
      errflg++;
      break;
    default:
      break;
    }
  }

  /* if no command given, and help asked for, print the list of commands.
     Otherwise, we'll pass the request for help onto the underlying command */
  if(errflg || (helpflg && optind >= argc)) {
    print_usage (argv[0]);
    exit (1);
  }

  if(optind >= argc) {
    print_usage (argv[0]);
    exit (1);
  }

  /* now work out which lcg command it is */
  if (strcmp (argv[optind], "aa") == 0
      || strcmp (argv[optind], "addAlias") == 0) {
    sprintf (cmd, "lcg-aa");
  } else if (strcmp (argv[optind], "ac") == 0 
	     || strcmp (argv[optind], "getAccessCost") == 0) {
    sprintf (cmd, "UNSUP");
  } else if (strcmp (argv[optind], "bcr") == 0 
	     || strcmp (argv[optind], "bulkCopyAndRegister") == 0) {
    sprintf (cmd, "UNSUP");
  } else if (strcmp (argv[optind], "cp") == 0 
	     || strcmp (argv[optind], "copyFile") == 0) {
    sprintf (cmd, "lcg-cp");
  } else if (strcmp (argv[optind], "cr") == 0 
	     || strcmp (argv[optind], "creg") == 0
      || strcmp (argv[optind], "copyAndRegisterFile") == 0) {
    sprintf (cmd, "lcg-cr");
  } else if (strcmp (argv[optind], "del") == 0 
	     || strcmp (argv[optind], "deleteFile") == 0) {
    sprintf (cmd, "lcg-del");
  } else if (strcmp (argv[optind], "gbf") == 0 
	     || strcmp (argv[optind], "getBestFile") == 0) {
    sprintf (cmd, "UNSUP");
  } else if (strcmp (argv[optind], "getVersion") == 0) {
    sprintf (cmd, "UNSUP");
  } else if (strcmp (argv[optind], "gt") == 0 
	     || strcmp (argv[optind], "getTurl") == 0) {
    sprintf (cmd, "lcg-gt");
  } else if (strcmp (argv[optind], "lbf") == 0 
	     || strcmp (argv[optind], "listBestFile") == 0) {
    sprintf (cmd, "UNSUP");
  } else if (strcmp (argv[optind], "lg") == 0
	     || strcmp (argv[optind], "listGUID") == 0) {
    sprintf (cmd, "lcg-lg");
  } else if (strcmp (argv[optind], "lr") == 0 
	     || strcmp (argv[optind], "listReplicas") == 0) {
    sprintf (cmd, "lcg-lr");
  } else if (strcmp (argv[optind], "ls") == 0
	     || strcmp (argv[optind], "list") == 0) {
    sprintf (cmd, "UNSUP");
  } else if (strcmp (argv[optind], "ra") == 0
	     || strcmp (argv[optind], "removeAlias") == 0) {
    sprintf (cmd, "lcg-ra");
  } else if (strcmp (argv[optind], "rep") == 0
	     || strcmp (argv[optind], "replicateFile") == 0) {
    sprintf (cmd, "lcg-rep");
  } else if (strcmp (argv[optind], "rf") == 0
	     || strcmp (argv[optind], "registerFile") == 0) {
    sprintf (cmd, "lcg-rf");
  } else if (strcmp (argv[optind], "rg") == 0
	     || strcmp (argv[optind], "registerGUID") == 0) {
    sprintf (cmd, "UNSUP");
  } else if (strcmp (argv[optind], "srm") == 0) {
    sprintf (cmd, "UNSUP");
  } else if (strcmp (argv[optind], "uf") == 0
	     || strcmp (argv[optind], "unregisterFile") == 0) {
    sprintf (cmd, "lcg-uf");
  } else if (strcmp (argv[optind], "pi") == 0
	     || strcmp (argv[optind], "printInfo") == 0) {
    sprintf (cmd, "UNSUP");
  } else if (strcmp (argv[optind], "validate") == 0) {
    sprintf (cmd, "UNSUP");
  } else{
    cmd[0] = '\0';
  }

  if(strcmp ("", cmd) == 0) {
    fprintf (stderr, "Error : No such command : %s\n" , argv[optind]);
    print_usage (argv[0]);
    exit (1);
  } else if (strcmp ("UNSUP", cmd) == 0) {
    fprintf (stderr, "Error : Unsupported command : %s\n", argv[optind]);
    print_usage (argv[0]);
    exit (1);
  }
  args[0] = cmd;

  if (vo_arg == NULL) {
     vo_arg = getenv("LCG_GFAL_VO");
     if (!vo_arg) {
	fprintf(stderr, "VO not specified and LCG_GFAL_VO not set.\n");
        exit(1);
       }
   }

  /* check for required default SE for lcg-cr & lcg-rep */
  if((strcmp("lcg-cr", cmd) == 0 || strcmp("lcg-rep", cmd) == 0) &&
     !has_dest) {
    /* try and get it from env variable */
    if(vo_arg != NULL) {
            if(strlen(vo_arg) + 15 >= 64) {
                    errno = ENAMETOOLONG;
                fprintf(stderr,"VO Name too long : %s\n", vo_arg);
                exit (1);
            }
            sprintf(se_env, "VO_%s_DEFAULT_SE", vo_arg);
            for(i = 3; i < 3 + strlen(vo_arg); ++i) {
                    if (se_env[i] == '.') 
                            se_env[i] = '_';
                    else
                            se_env[i] = toupper(se_env[i]);
            }
            if ((default_se = getenv(se_env)) == NULL) {
                    fprintf (stderr, "Error : No destination SE specified, and no %s in environment\n", se_env);
                    exit (1);
            }
            add_optarg("-d", default_se);
    } 
  }

  /* and finally add the optional arguments, if there are any */
  if(idx + (argc - optind) > size) {
      size += (argc - optind);
      if((p = (char**) realloc (args, size * sizeof(char*))) == NULL) {
	free (args);
	perror ("main");
	exit (1);
      }
      args = p;
  }

  for(i = optind + 1; i < argc; i++) {
    args[idx++] = argv[i];
  }
  args[idx++] = NULL;

  if(verbose) {
    fprintf (stderr, "Running command : ");
    for (p = args; *p != NULL; p++) {
      fprintf (stderr, "%s ", *p);
    }
    fprintf (stderr, "\n");
  }

  if(execvp (args[0], args) < 0) {
    perror ("lcg-replica-manager");
    exit (1);
  }

}


