/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef lint
static char sccsid[] = "@(#)$RCSfile: lcg_get_checksum.c,v $ $Revision: 1.2 $ $Date: 2009/04/08 14:23:45 $ CERN Remi Mollon";
#endif /* not lint */

/*	lcg_get_checksum - get file checksum */
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "gfal_api.h"
#include "lcg_util.h"


int
lcg_get_checksum (int nbfiles, char **files, enum se_type defaulttype, enum se_type setype, int nobdii,
		int force, char *vo, enum gfal_cksm_type cksmtype, lcg_checksum **checksums, char *errbuf, int errbufsz)
{
	int rc, i, sav_errno = 0;
	gfal_request req = NULL;
	gfal_internal gfal = NULL;
	char *protocols[] = {"gsiftp", NULL};
	int nberrors = 0;
	char errmsg[GFAL_ERRMSG_LEN];

	if (nbfiles < 1 || files == NULL || checksums == NULL) {
		errno = EFAULT;
		return (-1);
	}
	if (vo && vo[0]) {
		if (strlen (vo) > GFAL_VO_MAXLEN) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "VO name too long.");
			errno = ENAMETOOLONG;
			return (-1);
		}
		if (gfal_set_vo (vo) < 0)
			return (-1);
	}
	
	/* if no algo is specified, use the default one (in gfal_constants.h) */
	if (cksmtype == GFAL_CKSM_NONE)
		cksmtype = GFAL_CKSM_DEFAULT;

	if (check_surls (nbfiles, files, errbuf, errbufsz) < 0)
		return (-1);

	gfal_set_nobdii (nobdii);
	gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "VO name: %s", gfal_get_vo (errmsg, GFAL_ERRMSG_LEN));
	gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_WARN, "Checksum type: %s", gfal_get_checksum_algname (cksmtype));

	if (strncmp (files[0], "srm:", 4) == 0 || strncmp (files[0], "sfn:", 4) == 0) {
		if ((req = gfal_request_new ()) == NULL)
			return (-1);
		req->nbfiles = nbfiles;
		req->surls = files;
		req->defaultsetype = defaulttype;
		req->setype = setype;
		req->no_bdii_check = nobdii;
		req->protocols = protocols;

		rc = gfal_init (req, &gfal, errbuf, errbufsz);
		free (req);

		if (rc < 0)
			return (-1);

		rc = lcg_get_checksum_surls (gfal, force, cksmtype, checksums, errbuf, errbufsz);

		sav_errno = errno;
		gfal_internal_free (gfal);
        gfal = NULL;
		errno = sav_errno;
		return (rc);
	}

	if (strncmp (files[0], "lfn:", 4) == 0 || strncmp (files[0], "guid:", 5) == 0) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "lfn: and guid: URLs aren't supported (replicas may have different checksums)");
		errno = EPROTONOSUPPORT;
		return (-1);
	}

	/* else we consider files as turls and checksum is computed by reading the files (except gridftp) */
	if ((*checksums = (lcg_checksum *) calloc (nbfiles, sizeof (lcg_checksum))) == NULL)
		return (-1);

	for (i = 0; i < nbfiles; ++i) {
		char errmsg[GFAL_ERRMSG_LEN];

		(*checksums)[i].file = strdup (files[i]);

		if (strncmp (files[i], "gsiftp:", 7) == 0) {
			rc = gridftp_checksum (files[i], cksmtype, &((*checksums)[i].checksum), 0,
					errmsg, GFAL_ERRMSG_LEN, gfal_get_timeout_sendreceive ());

			if (rc < 0 || (*checksums)[i].checksum == NULL) {
				++nberrors;
				(*checksums)[i].status = errno > 0 ? errno : ECOMM;
				if (errmsg[0])
					(*checksums)[i].explanation = strdup(errmsg);
			}
		}
		else if (strncmp (files[i], "file:", 5) == 0) {
			errno = 0;
			errmsg[0] = 0;
			(*checksums)[i].checksum = lcg_compute_checksum (files[i], cksmtype, errmsg, GFAL_ERRMSG_LEN);
			if ((*checksums)[i].checksum == NULL) {
				++nberrors;
				(*checksums)[i].status = errno > 0 ? errno : ECOMM;
				if (errmsg[0])
					(*checksums)[i].explanation = strdup (errmsg);
			}
		}
		else {
			(*checksums)[i].status = EPROTONOSUPPORT;
		}
	}

	errno = 0;
	return (nberrors > 0 ? -1 : 0);
}

int
lcg_get_checksum_surls (gfal_internal gfal, int force, enum gfal_cksm_type cksmtype,
		lcg_checksum **checksums, char *errbuf, int errbufsz)
{
	int n, rc, i, j, sav_errno = 0;
	gfal_filestatus *filestatuses = NULL;
	char *protocols[] = {"gsiftp", NULL};
	int nb_remaining_surls = 0, nb_tmp_surls = 0, nberrors = 0;
	char **remaining_surls = NULL, **tmp_surls = NULL;
	int pos_checksums = 0;
	char errmsg[GFAL_ERRMSG_LEN];

	if (gfal == NULL || checksums == NULL) {
		errno = EFAULT;
		return (-1);
	}

	*checksums = (lcg_checksum *) calloc (gfal->nbfiles, sizeof (lcg_checksum));
	remaining_surls = (char **) calloc (gfal->nbfiles, sizeof (char **));
	if (*checksums == NULL || remaining_surls == NULL) {
		sav_errno = errno;
		if (*checksums) {
			free (*checksums);
			checksums = NULL;
		}
		if (remaining_surls) free (remaining_surls);
		errno = sav_errno;
		return (-1);
	}

	/* First try to get checksum with srmLs, if SE type is srmv2 */
	if (!force && gfal->setype == TYPE_SRMv2) {
		if (gfal_ls (gfal, errbuf, errbufsz) < 0) {
			free (*checksums); *checksums = NULL;
			free (remaining_surls);
			return (-1);
		}

		if ((n = gfal_get_results (gfal, &filestatuses)) < 0 || filestatuses == NULL) {
			free (*checksums); *checksums = NULL;
			free (remaining_surls);
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "Internal error");
			return (-1);
		}

		for (i = 0; i < gfal->nbfiles; ++i) {
			/* file names in results may be only parts of SURLs */
			if (!strstr (gfal->surls[i], filestatuses[i].surl)) {
				for (j = 0; j < gfal->nbfiles && !strstr (gfal->surls[j], filestatuses[i].surl); ++j) ;
				if (j >= gfal->nbfiles)
					continue;
			}
			else
				j = i;

			if (filestatuses[i].status != 0) {
				(*checksums)[pos_checksums].file = strdup(gfal->surls[j]);
				(*checksums)[pos_checksums].status = filestatuses[i].status;
				if (filestatuses[i].explanation)
					(*checksums)[pos_checksums].explanation = strdup(filestatuses[i].explanation);

				++pos_checksums;
			}
			else if (filestatuses[i].checksum && filestatuses[i].checksumtype) {
				if (strcasecmp (gfal_get_checksum_algname (cksmtype), filestatuses[i].checksumtype) == 0) {
					(*checksums)[pos_checksums].file = strdup(gfal->surls[j]);
					(*checksums)[pos_checksums].checksum = strdup (filestatuses[i].checksum);
					++pos_checksums;
				}
				else {
					gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_INFO, "[INFO] Checksum from srmLs: %s (%s)\t%s",
							filestatuses[i].checksum, filestatuses[i].checksumtype, filestatuses[i].surl);
					remaining_surls[nb_remaining_surls++] = gfal->surls[j];
				}
			}
			else {
				gfal_errmsg (NULL, 0, GFAL_ERRLEVEL_INFO, "[INFO] No checksum from srmLs: %s", filestatuses[i].surl);
				remaining_surls[nb_remaining_surls++] = gfal->surls[j];
			}
		}
	}
	else {
		/* copy of all surls in remaining_surls */
		for (i = 0; i < gfal->nbfiles; ++i)
			remaining_surls[nb_remaining_surls++] = gfal->surls[i];
	}

	if (nb_remaining_surls > 0) {
		/* checksum isn't available from srmLs for all SURLs...
		 * then get the gsiftp surl and ask for checksum with CKSM gridftp command */

		/* Replace the list of surls by the list of remaining surls
		 * keeping the same object (same endpoint, parameters, ...) */
		nb_tmp_surls = gfal->nbfiles;
		tmp_surls = gfal->surls;
		gfal->nbfiles = nb_remaining_surls;
		gfal->surls = remaining_surls;
		gfal->protocols = protocols;
		gfal->oflag = 0;

		if (gfal_turlsfromsurls (gfal, errbuf, errbufsz) < 0) {
			free (*checksums); *checksums = NULL;
			free (remaining_surls);
			return (-1);
		}

		if ((n = gfal_get_results (gfal, &filestatuses)) < 0 || filestatuses == NULL) {
			free (*checksums); *checksums = NULL;
			free (remaining_surls);
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "Internal error");
			return (-1);
		}

		for (i = 0; i < nb_remaining_surls; ++i) {
			(*checksums)[pos_checksums].file = strdup(filestatuses[i].surl);
			errmsg[0] = 0;

			if (filestatuses[i].status != 0) {
				(*checksums)[pos_checksums].status = filestatuses[i].status;
				if (filestatuses[i].explanation)
					(*checksums)[pos_checksums].explanation = strdup(filestatuses[i].explanation);

				++pos_checksums;
				++nberrors;
				continue;
			}

			rc = gridftp_checksum (filestatuses[i].turl, cksmtype, &((*checksums)[pos_checksums].checksum), 0,
					errmsg, GFAL_ERRMSG_LEN, gfal_get_timeout_sendreceive ());

			if (rc < 0 || (*checksums)[pos_checksums].checksum == NULL) {
				++nberrors;
				(*checksums)[pos_checksums].status = errno > 0 ? errno : ECOMM;
				if (errmsg[0])
					(*checksums)[pos_checksums].explanation = strdup(errmsg);
			}

			++pos_checksums;
		}

		gfal_release (gfal, errbuf, errbufsz);
		gfal->nbfiles = nb_tmp_surls;
		gfal->surls = tmp_surls;
	}

	free (remaining_surls);
	errno = 0;
	return (nberrors > 0 ? -1 : 0);
}
