/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifdef lint
static char sccsid[] = "@(#)$RCSfile: lcg_stmd.c,v $ $Revision: 1.7 $ $Date: 2009/04/02 15:38:51 $ CERN Remi Mollon";
#endif /* not lint */

/* lcg_stmd - get space tokens associated to a space token description, and their metadata */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include "gfal_api.h"
#include "lcg_util.h"
#include "gfal_srm_ifce.h"

int lcg_stmd
(
    const char *spacetokendesc,
    const char *endpoint,
    int *nbtokens,
    gfal_spacemd **spacemd,
    int nobdii,
    int timeout,
    char *errbuf,
    int errbufsz
)
{
	int i, ret;
	char *srmv2_endpoint = NULL;
	char **se_types = NULL, **se_endpoints = NULL, **spacetokens = NULL;
	struct srm_context context;
    struct srm_getspacetokens_input spt_input = {NULL};
    struct srm_getspacetokens_output spt_output = {0, NULL};
    struct srm_getspacemd_input smd_input = {0, NULL};
    struct srm_spacemd* smd_output = NULL;


	if (!spacetokendesc || !endpoint || !nbtokens || !spacemd) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "lcg_stmd: Invalid arguments");
		errno = EINVAL;
		return (-1);
	}

	gfal_set_verbose (0);

	if (timeout > 0) {
		// for backward compatibility...
		gfal_set_timeout_connect (timeout);
		gfal_set_timeout_sendreceive (timeout);
		gfal_set_timeout_bdii (timeout);
		gfal_set_timeout_srm (timeout);
	}

	if (!nobdii) {
		if (setypesandendpoints (endpoint, &se_types, &se_endpoints, errbuf, errbufsz) < 0 ||
				se_types == NULL || se_endpoints == NULL)
			return (-1);

		for (i = 0; se_types[i]; ++i) {
			if (srmv2_endpoint == NULL && strcmp (se_types[i], "srm_v2") == 0)
				srmv2_endpoint = se_endpoints[i];
			else
				free (se_endpoints[i]);

			free (se_types[i]);
		}

		free (se_types);
		free (se_endpoints);

		if (!srmv2_endpoint) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s: No matching SRMv2 endpoint in BDII", endpoint);
			errno = EINVAL;
			return (-1);
		}
	}
	else {
		if(strncmp (endpoint, "httpg://", 8)) {
			asprintf (&srmv2_endpoint, "httpg://%s", endpoint);
			if (srmv2_endpoint == NULL)
				return (-1);
		} else
			srmv2_endpoint = (char *) endpoint;
	}

    srm_context_init(&context, srmv2_endpoint, errbuf, errbufsz, 1);
    spt_input.spacetokendesc = (char*) spacetokendesc;
    ret = srm_getspacetokens (&context, &spt_input, &spt_output);
    *nbtokens = spt_output.nbtokens;
    spacetokens = spt_output.spacetokens;

	if (ret < 0 || spacetokens == NULL || *nbtokens < 1) {
		if (srmv2_endpoint != endpoint) free (srmv2_endpoint);
		errno = ret == 0 ? ENOMEM : errno;
		return (-1);
	}


    smd_input.nbtokens = *nbtokens;
    smd_input.spacetokens = spacetokens;
    ret = srm_getspacemd (&context, &smd_input, &smd_output);

	for (i = 0; i < *nbtokens; ++i)
		if (spacetokens[i]) free (spacetokens[i]);
	free (spacetokens);
	if (srmv2_endpoint != endpoint) free (srmv2_endpoint);

	*spacemd= smd_output; // transmit the result
	
	if (ret < 0 || *spacemd == NULL) {
		errno = ret == 0 ? ENOMEM : errno;
		return (-1);
	}

	return (0);
}
