/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * @(#)$RCSfile: lcg_util_common.c,v $ $Revision: 1.10 $ $Date: 2009/11/18 13:53:33 $ CERN Remi Mollon
 */

#define LCG_READBUFLEN 4096
#define LCG_DEFSURLTABSIZE 256

#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <dlfcn.h>
#include <openssl/evp.h>
#include "gfal_api.h"
#include "gfal_internals.h"
#include "lcg_util.h"

#define GFAL_CKSM_BUFSIZE 1048576	// = 1 MB

/* the version should be set by a "define" at the makefile level */
static const char lcgutilversion[] = VERSION;

/* zlib is dynamically loaded (if needed) */
static void *zlib_handle = NULL;
static unsigned long (*zlib_crc32) (unsigned long, const char *, unsigned int) = NULL;
static unsigned long (*zlib_adler32) (unsigned long, const char *, unsigned int) = NULL;

static const char *se_type_string[4] = { "NONE", "SRMv1", "SRMv2", "Classic SE" };

/* default callback function for a gsiftp tranfer */
static copyfile_callback_t transfer_cb = NULL;

const char *
lcg_util_version () {
	return lcgutilversion;
}

const char *
lcg_se_type_string (enum se_type setype) {
	return (se_type_string[(int) setype]);
}

int
import_surls_from_file (const char *filename, char ***surls, char *errbuf, int errbufsz) {
	int fd;
	int pos = 0, newpos = 0, nbsurls = 0, surls_size = 0;
	ssize_t size = 0;
	char buf[LCG_READBUFLEN + 1];
	char *endofline = NULL;
	char **newsurls = NULL;

	if (filename == NULL || surls == NULL) {
		errno = EINVAL;
		return (-1);
	}

	fd = open (filename, O_RDONLY);
	if (fd == -1) return (-1);

	*surls = (char **) calloc (LCG_DEFSURLTABSIZE, sizeof (char *));
	surls_size = LCG_DEFSURLTABSIZE;
	if (*surls == NULL) {
		close (fd);
		return (-1);
	}

	while ((size = read (fd, buf + pos, LCG_READBUFLEN - pos)) > 0) {
		buf[pos + size] = 0;
		pos = 0;

		while ((endofline = strchr (buf + pos, '\n')) != NULL) {
			if (buf[pos] != '#') {
				/* If it starts with '#', it is commented, so the line is ignored */
				*endofline = '\0';

				if (nbsurls == surls_size - 1) {
					/* 'surls' is full, it is reallocated to a double-sized tab */
					surls_size *= 2;
					newsurls = (char **) realloc (*surls, surls_size * sizeof (char *));
					if (newsurls == NULL) {
						int i;

						for (i = 0; i < nbsurls; ++i)
							free ((*surls)[i]);

						free (surls);
						surls = NULL;
						close (fd);
						return (-1);
					}

					*surls = newsurls;
				}

				(*surls)[nbsurls] = strdup (buf + pos);
				if ((*surls)[nbsurls] == NULL) {
					/* there was a problem, so all is freed... */
					int i;

					for (i = 0; i < nbsurls; ++i)
						free ((*surls)[i]);

					free (*surls);
					surls = NULL;
					close (fd);
					return (-1);
				}

				++nbsurls;
			}

			pos = (endofline - buf) + 1;
		}

		/* The remaining characters are copied to the beginning of the buffer
		 * and the 'pos' is set to the next position in the buffer */
		newpos = LCG_READBUFLEN - pos;
		memmove (buf, buf + pos, newpos * sizeof (char));
		pos = newpos;
	}

	if (nbsurls + 1 < surls_size) {
		/* Decrease the size of req->surls to the minimum */
		newsurls = (char **) realloc (*surls, (nbsurls + 1) * sizeof (char*));
		if (newsurls == NULL) {
			int i;

			for (i = 0; i < nbsurls; ++i)
				free ((*surls)[i]);

			free (surls);
			surls = NULL;
			close (fd);
			return (-1);
		}

		newsurls[nbsurls] = NULL;
		*surls = newsurls;
	}

	close (fd);
	return (nbsurls);
}

int
check_surls (int nbsurls, char **surls, char *errbuf, int errbufsz) {
	char *p1 = NULL, *p2 = NULL;
	char beginning[GFAL_PATH_MAXLEN];
	int i, len;

	beginning[0] = 0;

	if (nbsurls < 1 || surls == NULL) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "You must specify SURLs");
		errno = EINVAL;
		return (-1);
	}

	for (i = 0; i < nbsurls; ++i) {
		if ((p1 = strchr (surls[i], ':')) == NULL) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s: Invalid SURL", surls[i]);
			errno = EINVAL;
			return (-1);
		}
		if (beginning[0] != 0 && strncmp (beginning, surls[i], p1 - surls[i] + 1) != 0) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "All SURLs must have the same type (lfn:, srm:, ...)");
			errno = EINVAL;
			return (-1);
		}
		if (strncmp (surls[i], "srm", p1 - surls[i]) == 0) {
			if (*(p1 + 1) != '/' || *(p1 + 2) != '/' ||
					((p2 = strchr (p1 + 3, '?')) == NULL &&
					 (p2 = strchr (p1 + 3, '/')) == NULL)) {
				gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "%s: Invalid SURL", surls[i]);
				errno = EINVAL;
				return (-1);
			}
			if (beginning[0] != 0) {
				int len_tmp = p2 - surls[i];

				/* The 1st SURL must have the most complete endpoint
				 * Others' endpoint can be only a part of it
				 * eg. 1. srm://lxdpm102.cern.ch:8446/srm/managerv2?SFN=/dpm/...
				 *        srm://lxdpm102.cern.ch/dpm/... */
				if (strncmp (surls[i], beginning, len_tmp) != 0 ||
						(beginning[len_tmp] != '/' &&
						 beginning[len_tmp] != ':' &&
						 beginning[len_tmp] != '?')) {
					gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "All SURLs must be from the same SE endpoint");
					errno = EINVAL;
					return (-1);
				}
			}
			else {
				len = p2 - surls[i] + 1;
				snprintf (beginning, len + 1, "%s", surls[i]);
			}
		}
		else { /* for lfn:, guid:, sfn:, gsiftp:, file:, ... */
			if (beginning[0] == 0) {
				len = p1 - surls[i] + 1;
				snprintf (beginning, len + 1, "%s", surls[i]);
			}
		}
	}

	return (0);
}

void
lcg_set_transfer_callback (copyfile_callback_t cb) {
	transfer_cb = cb;
}

copyfile_callback_t
lcg_get_transfer_callback () {
	return (transfer_cb);
}

char *
lcg_compute_checksum (const char *file, enum gfal_cksm_type cksmtype, char *errbuf, int errbufsz)
{
	unsigned long cksm = 0;
	EVP_MD_CTX evpctx;
	int fd = -1;
	unsigned int nbread = 0;
	char buffer[GFAL_CKSM_BUFSIZE];
	char checksum[2 * EVP_MAX_MD_SIZE + 1];
	unsigned char cksm_raw[EVP_MAX_MD_SIZE + 1];
	int cksmsize = 0;

	if (file == NULL || cksmtype == GFAL_CKSM_NONE) {
		gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[LCG-UTIL][compute_checksum][] Invalid arguments");
		errno = EINVAL;
		return (NULL);
	}

	memset (checksum, 0, (2 * EVP_MAX_MD_SIZE + 1) * sizeof (char));

	if (cksmtype == GFAL_CKSM_CRC32 || cksmtype == GFAL_CKSM_ADLER32) {
		zlib_handle = dlopen ("libz.so", RTLD_LAZY);
		if (zlib_handle == NULL) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[LCG-UTIL][compute_checksum][] CRC32/ADLER32 need zlib!");
			errno = ELIBACC;
			return (NULL);
		}

		zlib_crc32 = (unsigned long (*) (unsigned long, const char *, unsigned int)) dlsym (zlib_handle, "crc32");
		zlib_adler32 = (unsigned long (*) (unsigned long, const char *, unsigned int)) dlsym (zlib_handle, "adler32");
		if (zlib_crc32 == NULL || zlib_adler32 == NULL) {
			gfal_errmsg (errbuf, errbufsz, GFAL_ERRLEVEL_ERROR, "[LCG-UTIL][compute_checksum][] invalid zlib library!");
			errno = ELIBBAD;
			return (NULL);
		}
	}

	fd = gfal_open (file, O_RDONLY, 0);
	if (fd < 0)
		return (NULL);

	if (cksmtype == GFAL_CKSM_ADLER32)
		cksm = zlib_adler32 (0L, 0, 0);
	else if (cksmtype == GFAL_CKSM_CRC32)
		cksm = zlib_crc32 (0L, 0, 0);
	else {
		if (cksmtype == GFAL_CKSM_MD5)
			EVP_DigestInit (&evpctx, EVP_md5());
		else
			EVP_DigestInit (&evpctx, EVP_sha1());
	}

	while ((nbread = (unsigned int) gfal_read (fd, buffer, GFAL_CKSM_BUFSIZE)) > 0) {
		if (cksmtype == GFAL_CKSM_ADLER32)
			cksm = zlib_adler32 (cksm, buffer, nbread);
		else if (cksmtype == GFAL_CKSM_CRC32)
			cksm = zlib_crc32 (cksm, buffer, nbread);
		else
			EVP_DigestUpdate(&evpctx, buffer, (size_t) nbread);
	}

	gfal_close (fd);

	if (nbread < 0)
		return (NULL);

	if (cksmtype == GFAL_CKSM_ADLER32 || cksmtype == GFAL_CKSM_CRC32) {
		sprintf(checksum, "%08x", (unsigned int) cksm);
	} else {
		char *p;
		int i;

		EVP_DigestFinal (&evpctx, cksm_raw, &cksmsize);

		for (i = 0, p = checksum; i < cksmsize; ++i, p = p + 2)
			sprintf(p, "%02x", cksm_raw[i]);
	}

	return (strdup (checksum));
}

void lcg_free_string_array(char** s)
{
    if (s) {
        int i = 0;

        while (s[i]) {
            free(s[i]);
	    s[i]=NULL;
            ++i;
        }
       if(*s!=NULL){
        free(s);
	s=NULL;
	}
    }
}


char ** lcg_tokenize(const char *str, const char *sep)
{
    char **tokens = NULL;
    char *strtmp = NULL;
    char *word = NULL;
    int n = 0;
    int i = 0;
    /* No tokens for null strings. */
    if (str == NULL || sep == NULL)
        return NULL;

    strtmp = (char *)malloc(sizeof(char *) * strlen(str) + 1);
    strcpy(strtmp, str);
    /* Calculate number of tokens. */
    for (word = strtok(strtmp, sep);
         word;
         word = strtok(NULL, sep))
    {
        ++n;
    }
    /* Allocate memory for the pointer to each token. */
    tokens = (char **)malloc(sizeof(char *) * n + 1);
    strcpy(strtmp, str);
    /* Allocate memory for each token. */
    for (word = strtok(strtmp, sep);
         word;
         word = strtok(NULL, sep))
    {
        tokens[i] = (char *)malloc(sizeof(char) * (strlen(word) + 1));
        strcpy(tokens[i], word);
        i++;
    }
    /* The closing NULL (indicating the end of the list) */
    tokens[i] = NULL;

    free(strtmp);
    return tokens;
}


void lcgutil_str_tolower (char* str)
{
    int i = 0;

    for (i = 0; str[i] != 0; ++i)
    {
        char c = tolower( str[i] );
        str[i] = c;
    }
}


